/* 
 *  Mezzanine - an overhead visual object tracker.
 *  Copyright (C) Andrew Howard 2002
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
/***************************************************************************
 * Desc: Mezzanine calibration tool.
 * Author: Andrew Howard
 * Date: 28 Mar 2002
 * CVS: $Id: mezzcal.c,v 1.3 2002/05/06 05:33:00 inspectorg Exp $
 ***************************************************************************/

#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include "opt.h"
#include "mezzcal.h"

// Set flag to 1 to force program to quit
static int quit = 0;

// Handle quit signals
void sig_quit(int signum)
{
  quit = 1;
}


// Main
int main(int argc, char **argv)
{
  mezz_mmap_t *mmap;
  int rate;
  rtk_app_t *app;  
  imagewnd_t *imagewnd;
  colorwnd_t *colorwnd;
  tablewnd_t *tablewnd;

  printf("MezzCal %s\n", VERSION);

  // Initialise rtk lib
  rtk_init(&argc, &argv);
    
  // Register signal handlers
  signal(SIGINT, sig_quit);
  signal(SIGQUIT, sig_quit);
  
  // Load program options
  if (opt_init(argc, argv, NULL) != 0)
    return -1;

  // Pick out some important program options
  rate = opt_get_int("gui", "rate", 5);
  
  // Initialise the IPC
  if (mezz_init(0) < 0)
    return -1;
  mmap = mezz_mmap();

  // Make sure the IPC is running before we do anything else
  mezz_wait_event();
  
  // Enable the vision stuff in the ipc
  mmap->calibrate++;

  // Create gui
  app = rtk_app_create();

  // Create a window for the raw image
  imagewnd = imagewnd_init(app, mmap);
  if (!imagewnd)
    return -1;

  // Create a windiow for the YUV color space
  colorwnd = colorwnd_init(app, mmap);
  if (!colorwnd)
    return -1;

  // Create the property table
  tablewnd = tablewnd_init(app, mmap);
  if (!tablewnd)
    return -1;

  // Initialise interfaces to different parts of the
  // vision system.  
  if (image_init(imagewnd, mmap) != 0)
    return -1;
  if (mask_init(imagewnd, mmap) != 0)
    return -1;
  if (sample_init(imagewnd, colorwnd, mmap) != 0)
    return -1;
  if (classify_init(colorwnd, mmap) != 0)
    return -1;
  if (blobfind_init(imagewnd, tablewnd, mmap) != 0)
    return -1;
  if (dewarp_init(imagewnd, mmap) != 0)
    return -1;
  if (ident_init(imagewnd, mmap) != 0)
    return -1;

  // Check to see if there were any unrecognized options
  opt_warn_unused();

  // Start the gui
  rtk_app_refresh_rate(app, rate);
  rtk_app_start(app);
  
  while (!quit)
  {
    mezz_wait_event();

    imagewnd_update(imagewnd); 
    
    image_update();
    mask_update();
    sample_update();
    classify_update();
    dewarp_update();
    blobfind_update();
    ident_update();
  }
  
  // Stop the gui
  rtk_app_stop(app);
  rtk_app_destroy(app);

  // Disable the vision stuff in the ipc
  mmap->calibrate--;

  // Finalise the IPC
  mezz_term(0);
  
  return 0;
}
