#include "Brain.h"
#include "Robot.h"
#include "PotentialMap.h"
#include "RoboMath.h"
#include "VisionClient.h"
#include "CommServer.h"
#include "Ball.h"
#include "Config.h"
#include "ActionTable.h"
#include <string.h>

VisionClient visionClient;
CommServer commServer;
Ball* ball;
Robot* robots;
Config config;
int state;
ActionTable actionTable;

PotentialMap potentialMaps[NUM_POTENTIAL_MAPS];

void initMaps();
void initObjects(Robot* robots, Ball* ball);
void generateStaticMaps(int direction);
void generateDynamicMaps();
PotentialMap generateRobotMaps(int robotID, double sigma);
PotentialMap generateUpfield(int upfieldDir);
PotentialMap generateCenter();
PotentialMap generateGoal(int dir);
PotentialMap generateBlocked();




int main(int argc, char* argv[]){
  int ManualIterationStep = 0;

  /***********************************
   * parse conmmand line args        *
   ***********************************/
  if (argc>1) {
    if (!strcmp(argv[1],"step")) {
      ManualIterationStep = 1;
    } 
    if (!strcmp(argv[1],"verbose")) {
      config.verbose = 1;
    } 
    if (argc>2){
      if (!strcmp(argv[2],"step")) {
	ManualIterationStep = 1;
      } 
      if (!strcmp(argv[2],"verbose")) {
	config.verbose = 1;
      } 
    }
  }
  
  /***********************************
   * initialization                  *
   ***********************************/
  robots = new Robot[8];
  ball = new Ball();
  
  initObjects(robots, ball);
  
  initMaps();

  visionClient.init(config.visHost, config.visPort);

  commServer.init(config.commPort);
  
  //direction / dir / whatever is hardcoded here, the way real programmers do it.
  generateStaticMaps(config.direction);
  //potentialMaps[GOAL].print();
 
 
  /***********************************
   * loop                            *
   ***********************************/
  while(1) {
    if(config.verbose){
      cout << "\n===============Next frame================\n";
      cout << "Requesting data...";
    }
    visionClient.getData(robots, ball);

    if(config.verbose){
      cout << "Got data.\n";
      ball->getPosition().print("  Ball position");
    }
    
    
    if(config.verbose)
      for(int i = 0; i < config.numOurRobots; i++)
	{
	  printf("  Robot %d:\t", i);
	  printf("ori: %lf\t", robots[i].getOrientationDouble());
	  robots[i].getPosition().print("   pos");
	  robots[i].getMotors().print("   motors");
	}
    
    
    
    generateDynamicMaps();
    //potentialMaps[ENEMY].print();
    //robots[1].getMap().print();
    
    if(config.verbose){
      printf("\n--------Generate Action Table---------\n");
    }
    
    actionTable.calculate(robots, potentialMaps, ball);
    
    if(config.verbose){
      actionTable.print();
      printf("\n-----------Assign Actions-------------\n");
    }
    actionTable.assignActions(robots);
    
    if(config.verbose){
      printf("\n-------------Send Data---------------\n");  
    }
    commServer.sendData(robots);
    
    if (ManualIterationStep == 1) getchar();
  }
  
  //endloop
  
  exit(0);
  
  //free pointers
  //delete[] robots;
  //delete ball;
}
  
void initObjects(Robot* robots, Ball* ball)
{
   
  for(int i = 0; i < config.numRobots; i++)
    {
      robots[i].init(config.fieldDimensions, config.numCells);
      robots[i].setName(config.names[i]);
      robots[i].setID(config.ids[i]);
    }

  ball->setName(config.ballName);
}

  

void initMaps() {

	for (int i = 0; i < NUM_POTENTIAL_MAPS; i++) {
	  potentialMaps[i] = PotentialMap(config.fieldDimensions, config.numCells);
	}
}

void generateStaticMaps(int direction)
{
  potentialMaps[UPFIELD] = generateUpfield( direction );
  
  potentialMaps[CENTER] = generateCenter();

  potentialMaps[GOAL] = generateGoal( direction );
  
}

void generateDynamicMaps()
{
	//update friendlies
	for (int i = 0; i < 4; i++) {
		robots[i].setMap(generateRobotMaps(i, config.friendlySigma));
	}
	//update enemies
	potentialMaps[ENEMY].clear();
	for (int i = 4; i < 8; i++)	{
		robots[i].setMap(generateRobotMaps(i, config.friendlySigma));
		potentialMaps[ENEMY] = potentialMaps[ENEMY] + robots[i].getMap();
	}
}

PotentialMap generateRobotMaps(int robotID, double sigma) 
{
	double value = 1, dist;
	PotentialMap map(config.fieldDimensions, config.numCells);
	PotentialMap map2(config.fieldDimensions, config.numCells);

    map = robots[robotID].getMap();
	for (int i = 0; i < config.numCells.x(); i++) {
		for (int j = 0; j < config.numCells.y(); j++) {
		  dist = (robots[robotID].getPosition() - map.cellToCoord(i, j)).magnitude();
		  value = gauss(dist, config.friendlySigma);
		  map.setCell(i, j, value);
		}
	}

	return map;
}

PotentialMap generateUpfield(int upfieldDir) {
	PotentialMap map(config.fieldDimensions, config.numCells);
	
	if (upfieldDir == POSITIVE) {
		for (int i = 0; i < (int)config.numCells.x(); i++) {
			for (int j = 0; j < (int)config.numCells.y(); j++) {
			  map.setCell(i,j, (double)j / config.numCells.y());
			}
		}
	}
	else if (upfieldDir == NEGATIVE) {
		for (int i = 0; i < config.numCells.x(); i++) {
			for (int j = 0; j < config.numCells.y(); j++) {
				map.setCell(i,j, (double)(config.numCells.y() - j) / config.numCells.y());
			}
		}
	}
	
	return map;
}

PotentialMap generateCenter() {
	PotentialMap map(config.fieldDimensions, config.numCells);
	for(int i = 0; i < config.numCells.x(); i++)
	{
		for(int j = 0; j < config.numCells.y(); j++)
		{
			map.setCell(i,j, fabs((double)(config.numCells.x()/2) - i) / (config.numCells.x()/2));
		}
	}

	return map;
}

PotentialMap generateGoal(int dir) {
  PotentialMap map(config.fieldDimensions, config.numCells);
  int startCell, endCell;
  int yValue;
  if(dir == POSITIVE)
    {
      yValue = (int)config.numCells.y() - 1;
    }
  else
    {
      yValue = 0;
    }

   startCell = (int)(map.coordToCell((config.fieldDimensions.x() - config.goalWidth)/2, 0)).x();
   endCell =  (int)(map.coordToCell((config.fieldDimensions.x() + config.goalWidth)/2, 0)).x();
 
  for(int i = startCell; i <= endCell; i++)
    {
      map.setCell(i, yValue, 1);
    }
  return map;

}

PotentialMap generateBlocked()
{
  /* NEEDS TO BE COMPLETED */
  PotentialMap map(config.fieldDimensions, config.numCells);
  Vector ballToRobot, posRadius, negRadius;
  int intersected;
  



  /*

  for(int i = 0; i < numCells.x; i++)
    {
      for(int j = 0; j < numCells.y; j++)
	{
	  intersected = 0;
	  ballToPointRay = map.cellToCoord(i, j) - ball;

	  for(int i = 0; i < NUM_ROBOTS; i++)
	    {
	      if(ballToPointRay.intersectsDisc(robot[i].getPosition(), config.robotRadius))
		{
		  intersected = 1;
		  break;
		}
	    }

	  if(intersected)
	    {
	      map.setCell(i, j, 0);
	    }
	  else
	    {
	      map.setCell(i, j, 1);
	    }
	}
    }

  */

  return map;
}

	  
